import pytest

from src.container import Container
from src.decorator import Depends
from src.integrations.celery import celery_task


def test_dummy_celery_injector():
    container = Container()

    class Dependency:
        def __init__(self):
            self.x = 1

    class Celery:
        def task(self, fn):
            return fn

    celery = Celery()

    @celery_task(celery, container)
    def my_function(dependency: Depends(Dependency)):
        return dependency.x

    assert my_function() == 1



"""create cow table

Revision ID: 04658d0dab42
Revises: 
Create Date: 2020-02-18 22:05:18.342280

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "04658d0dab42"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "cow",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("pregnant", sa.Boolean(), nullable=True),
        sa.Column("name", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )


def downgrade():
    op.drop_table("cow")



from functools import wraps
from typing import Any, Callable, Dict, Type, TypeVar, cast, get_type_hints

from .resolver import DependencyResolver


class DecoratorResolver:
    def __init__(self, resolver: DependencyResolver) -> None:
        self.__resolver = resolver

    def inject_function(self, fn: Callable) -> Callable:
        type_hints = get_type_hints(fn)
        injected_args: Dict[str, Any] = {}

        for parameter_name, parameter_dependency in type_hints.items():
            if parameter_name == "return" or not isinstance(parameter_dependency, ParameterDependence):
                continue

            injected_args[parameter_name] = self.__resolver.resolve(
                parameter_dependency.parameter_dependency
            ).get_instance()

        @wraps(fn)
        def wrapper(*args, **kwargs):
            return fn(*args, **{**injected_args, **kwargs})

        return wrapper


class ParameterDependence:
    def __init__(self, parameter_dependency: Type) -> None:
        self.parameter_dependency = parameter_dependency


T = TypeVar("T")


def Depends(parameter_type: Type[T]) -> T:
    return cast(T, ParameterDependence(parameter_type))



from typing import List

from .database import Database
from .models import Cow


class CowRepository:
    def __init__(self, database: Database):
        self.__session = database.get_session()

    def get_not_pregnant_cows(self, limit: int) -> List[Cow]:
        return self.__session.query(Cow).filter(Cow.pregnant == False).limit(limit).all()

    def inseminate_cow(self, cow: Cow) -> None:
        cow.pregnant = True
        self.__session.flush()

    def add_cow(self, name: str) -> Cow:
        cow = Cow(name=name, pregnant=False)
        self.__session.add(cow)
        return cow

    def commit(self) -> None:
        self.__session.commit()



class InseminatorError(Exception):
    pass


class ContainerError(InseminatorError):
    pass


class ContainerRegisterError(ContainerError):
    pass


class ResolverError(InseminatorError):
    pass



from inseminator import Container, Depends
from inseminator.integrations import celery_task

from .celery_factory import celery_factory
from .domain import DomainLogic

container = Container()
celery = container.resolve(celery_factory)


@celery_task(celery, container)
def inseminate(number_of_cows_to_inseminate: int, domain_logic: Depends(DomainLogic)) -> None:
    domain_logic.inseminate_cow(number_of_cows_to_inseminate)


@celery_task(celery, container)
def add_random_cows(number_of_cows_to_inseminate: int, domain_logic: Depends(DomainLogic)) -> None:
    domain_logic.add_random_cows(number_of_cows_to_inseminate)



from __future__ import annotations

from typing import Any, Callable, Dict, Type, TypeVar, Union

from .decorator import DecoratorResolver
from .dependency import Dependency, StaticDependency
from .exceptions import ContainerRegisterError
from .resolver import DependencyResolver
from .scoped_dict import ScopedDict

T = TypeVar("T")
Dependable = Union[Callable[..., T], Type[T]]


class Container:
    def __init__(self, parent_scoped_dict: ScopedDict[Dependable, Dependency] = None) -> None:
        self._container: ScopedDict[Dependable, Dependency] = ScopedDict(parent_scoped_dict)
        self._resolver = DependencyResolver(self._container)

    def register(
        self,
        dependency: Type[T],
        *,
        value: T = None,
        factory: Callable[..., T] = None,
        parameters: Dict[str, Dependable] = None,
    ) -> None:
        resolved_dependency: Dependency[T]

        if value is not None and factory is not None:
            raise ContainerRegisterError("Can't decide whether to use factory or value for dependency instantiation")

        if value is not None:
            resolved_dependency = StaticDependency(value)
        elif factory is not None:
            resolved_dependency = self._resolver.resolve(factory, parameters)
        else:
            resolved_dependency = self._resolver.resolve(dependency, parameters)

        self._container[dependency] = resolved_dependency

    def resolve(self, dependency: Type[T], **parameters: Dependable) -> T:
        if dependency not in self._container:
            self.register(dependency, parameters=parameters)

        return self._container[dependency].get_instance()

    def sub_container(self) -> Container:
        return Container(parent_scoped_dict=self._container)

    def inject(self, fn: Callable[..., T]) -> Callable[..., T]:
        decorator_resolver = DecoratorResolver(resolver=self._resolver)
        return decorator_resolver.inject_function(fn)



from celery import Celery

from .configuration import Configuration


def celery_factory(configuration: Configuration) -> Celery:
    return Celery("tasks", broker=configuration.redis_connection_string)



from inseminator import Container, Depends


class Dependency:
    def __init__(self):
        self.x = 1


container = Container()


@container.inject
def my_handler(input_value: int, dependency: Depends(Dependency)):
    return input_value + dependency.x



import requests


class RandomNameApiClient:
    def __init__(self) -> None:
        self.__url = "https://uinames.com/api/"

    def get_random_name(self) -> str:
        response = requests.get(self.__url, params={"gender": "female"})
        response.raise_for_status()

        return response.json()["name"]



import typing

import pytest

from src.container import Container
from src.exceptions import ContainerRegisterError


def test_basic_register_resolve():
    class TestDependency:
        def __init__(self):
            self.x = 1

    container = Container()
    container.register(TestDependency)

    assert container.resolve(TestDependency).x == 1


def test_multiple_basic_register_resolve():
    class TestDependency:
        def __init__(self):
            self.x = 1

    class AnotherTestDependency:
        def __init__(self):
            self.x = 2

    container = Container()
    container.register(TestDependency)
    container.register(AnotherTestDependency)

    assert container.resolve(TestDependency).x == 1
    assert container.resolve(AnotherTestDependency).x == 2


def test_dependency_without_registered_dependencies():
    class TestDependency:
        def __init__(self):
            self.x = 1

    container = Container()

    dependency_1 = container.resolve(TestDependency)
    dependency_2 = container.resolve(TestDependency)

    assert dependency_1.x == 1
    assert dependency_2.x == 1
    assert dependency_1 is dependency_2


def test_simple_hierarchy_depedency():
    class TestDependency:
        def __init__(self):
            self.x = 1

    class AnotherTestDependency:
        def __init__(self, dependency: TestDependency):
            self.x = dependency.x + 1

    container = Container()
    container.register(AnotherTestDependency)

    assert container.resolve(AnotherTestDependency).x == 2


def test_simple_hierarchy_depedency_without_registerd_dependency():
    class TestDependency:
        def __init__(self):
            self.x = 1

    class AnotherTestDependency:
        def __init__(self, dependency: TestDependency):
            self.x = dependency.x + 1

    container = Container()

    dependency_1 = container.resolve(AnotherTestDependency)
    dependency_2 = container.resolve(AnotherTestDependency)

    assert dependency_1 is dependency_2
    assert dependency_1.x == 2


def test_register_dependency_with_default_value():
    class TestDependency:
        def __init__(self, x):
            self.x = x

    class AnotherTestDependency:
        def __init__(self, dependency: TestDependency):
            self.x = dependency.x + 1

    container = Container()
    container.register(TestDependency, value=TestDependency(x=2))

    assert container.resolve(AnotherTestDependency).x == 3


def test_multiple_dependencies():
    class Dependency1:
        def __init__(self):
            self.x = 1

    class Dependency2:
        def __init__(self):
            self.x = 2

    class Dependency3:
        def __init__(self, dependency: Dependency1):
            self.x = dependency.x + 1

    class Dependency4:
        def __init__(self, dependency_3: Dependency3, dependency_2: Dependency2):
            self.x = dependency_3.x + dependency_2.x

    container = Container()

    assert container.resolve(Dependency4).x == 4


def test_register_dependency_by_factory():
    class Dependency:
        def __init__(self, x):
            self.x = x

    def dependency_factory() -> Dependency:
        return Dependency(1)

    class AnotherDependency:
        def __init__(self, dependency: Dependency):
            self.x = dependency.x

    container = Container()
    container.register(Dependency, factory=dependency_factory)

    assert container.resolve(AnotherDependency).x == 1


def test_value_and_factory_register_exception():
    class Dependency:
        def __init__(self, x):
            self.x = x

    def dependency_factory() -> Dependency:
        return Dependency(1)

    class AnotherDependency:
        def __init__(self, dependency: Dependency):
            self.x = dependency.x

    container = Container()

    with pytest.raises(ContainerRegisterError):
        container.register(Dependency, factory=dependency_factory, value=Dependency(2))



from typing import Any

from sqlalchemy import Boolean, Column, Integer, String
from sqlalchemy.ext.declarative import declarative_base

Base: Any = declarative_base()


class Cow(Base):
    __tablename__ = "cow"

    id = Column(Integer, primary_key=True)
    pregnant = Column(Boolean)
    name = Column(String)



from examples.basic import controller as basic_controller
from examples.basic_against_interface import controller as basic_against_interface_controller
from examples.decorator import my_handler
from examples.enforced_parameters import controller as enforced_parameters_controller


def test_basic():
    assert basic_controller.handler(1) == 2


def test_basic():
    assert basic_against_interface_controller.handler(1) == 2


def test_enforced_parameters():
    assert enforced_parameters_controller.handler(1) == 2


def test_decorator():
    assert my_handler(1) == 2



from typing import Callable, Protocol, TypeVar

from ..container import Container

T = TypeVar("T")


class Celery(Protocol):
    def task(self, fn: Callable[..., T]) -> T:
        raise NotImplementedError


def celery_task(celery: Celery, container: Container) -> Callable:
    def decorator(fn: Callable) -> Callable:
        return celery.task(container.inject(fn))

    return decorator



from typing import Any, Generic, Protocol, TypeVar

T = TypeVar("T", covariant=True)


class Dependency(Protocol, Generic[T]):
    def get_instance(self) -> T:
        raise NotImplementedError


class StaticDependency(Generic[T]):
    def __init__(self, instance: T) -> None:
        self.__instance = instance

    def get_instance(self) -> T:
        return self.__instance



import os


class Configuration:
    def __init__(self):
        self.connection_string = os.environ["DATABASE_CONNECTION_STRING"]
        self.redis_connection_string = os.environ["REDIS_CONNECTION_STRING"]



from typing import Protocol

import pytest

from src.container import Container
from src.exceptions import ResolverError


def test_sub_container():
    class Dependency(Protocol):
        x: int

    class Dependency1:
        x = 1

    class Dependency2:
        x = 2

    class Client:
        def __init__(self, dependency: Dependency) -> None:
            self.x = dependency.x

    container = Container()

    sub_container = container.sub_container()
    sub_container.register(Dependency, value=Dependency1())

    with pytest.raises(ResolverError) as e:
        container.resolve(Client)

    client1 = sub_container.resolve(Client)

    assert client1.x == 1

    container.register(Dependency, value=Dependency2())
    client2 = container.resolve(Client)

    assert client2.x == 2


def test_sub_container_2():
    class Dependency(Protocol):
        x: int

    class Dependency1:
        x = 1

    class Client:
        def __init__(self, dependency: Dependency) -> None:
            self.x = dependency.x

    container = Container()
    container.register(Dependency, value=Dependency1())

    sub_container = container.sub_container()
    client = sub_container.resolve(Client)

    assert client.x == 1


def test_resolving_with_parameters():
    class FirstDependency:
        x = 1

    class Dependency(Protocol):
        y: int

    class ConcreteDependency:
        y = 2

    class Client:
        def __init__(self, first_dependency: FirstDependency, second_dependency: Dependency) -> None:
            self.x = first_dependency.x
            self.y = second_dependency.y

        def add(self) -> int:
            return self.x + self.y

    container = Container()
    client = container.resolve(Client, second_dependency=ConcreteDependency)

    assert client.add() == 3


def test_resolving_with_wrong_parameters():
    class Dependency:
        x = 1

    class Client:
        def __init__(self, dependency: Dependency) -> None:
            self._dependency = dependency

    container = Container()

    with pytest.raises(ResolverError):
        client = container.resolve(Client, non_existing_parameter=Dependency())



from .cow_repository import CowRepository
from .models import Cow
from .random_name_api_client import RandomNameApiClient


class DomainLogic:
    def __init__(self, cow_repository: CowRepository, random_name_api_client: RandomNameApiClient) -> None:
        self.__cow_repository = cow_repository
        self.__random_name_api_client = random_name_api_client

    def inseminate_cow(self, number_of_cows_to_inseminate: int) -> None:
        print(f"Inseminating {number_of_cows_to_inseminate} cows.")
        cows = self.__cow_repository.get_not_pregnant_cows(number_of_cows_to_inseminate)

        for cow in cows:
            self.__cow_repository.inseminate_cow(cow)
            print(f"{cow.name} is now inseminated.")

        self.__cow_repository.commit()

    def add_random_cows(self, number_of_cows_to_add: int) -> None:
        for cow in range(number_of_cows_to_add):
            name = self.__random_name_api_client.get_random_name()
            cow = self.__cow_repository.add_cow(name=name)
            print(f"{cow.name} added to the farm.")

        self.__cow_repository.commit()



import inspect
from typing import Any, Callable, Dict, Optional, Protocol, Type, TypeVar, Union, cast, get_type_hints

from .dependency import Dependency, StaticDependency
from .exceptions import ResolverError
from .scoped_dict import ScopedDict

T = TypeVar("T")

Dependable = Union[Callable[..., T], Type[T]]


class DependencyResolver:
    def __init__(self, container: ScopedDict[Dependable, Dependency]) -> None:
        self._container = container

    def resolve(self, dependency: Dependable, parameters: Optional[Dict[str, Any]] = None) -> Dependency[T]:
        if dependency in self._container:
            return self._container[dependency]

        if inspect.isclass(dependency):
            if issubclass(cast(type, dependency), cast(type, Protocol)):
                raise ResolverError(f"Implementation for {dependency.__name__} protocol is not defined.")

            type_hints = get_type_hints(dependency.__init__)  # type: ignore
        elif inspect.isfunction(dependency):
            type_hints = get_type_hints(dependency)

        args: Dict[str, Any] = {}

        if parameters:
            parameter_names = type_hints.keys()

            for parameter_name, parameter_value in parameters.items():
                if parameter_name not in parameter_names:
                    raise ResolverError(f"Parameter {parameter_name} it not part of {dependency.__name__}'s signature.")

                args[parameter_name] = parameter_value

        for parameter_name, parameter_dependency in type_hints.items():
            if parameter_name == "return" or parameter_name in args:
                continue

            args[parameter_name] = self.resolve(parameter_dependency).get_instance()

        return StaticDependency(instance=dependency(**args))



from __future__ import annotations

from typing import Generic, Optional, TypeVar

K = TypeVar("K")
V = TypeVar("V")


class ScopedDict(dict, Generic[K, V]):
    def __init__(self, parent_dict: Optional[ScopedDict[K, V]] = None) -> None:
        self.__parent_dict = parent_dict

    def __getitem__(self, key: K) -> V:
        if not super().__contains__(key) and self.__parent_dict is not None:
            return self.__parent_dict[key]

        return super().__getitem__(key)

    def __contains__(self, key: object) -> bool:
        if not super().__contains__(key):
            if self.__parent_dict is not None:
                return key in self.__parent_dict

            return False

        return True



from src.container import Container
from src.decorator import Depends


def test_decorator_dependency():
    class Dependency:
        def __init__(self):
            self.x = 1

    container = Container()

    @container.inject
    def my_function(dependency: Depends(Dependency), value: int) -> int:
        return value + dependency.x

    assert my_function(value=1) == 2



#!/usr/bin/env python

import setuptools
from distutils.core import setup

with open("readme.md", "r") as fh:
    long_description = fh.read()

setup(
    name="inseminator",
    version="0.2",
    description="Python dependency injection library based on type hints",
    author="Milan Suk",
    author_email="Milansuk@email.com",
    url="https://www.github.com/sukovanej/inseminator/",
    package_dir={"inseminator": "src", "inseminator.integrations": "src/integrations"},
    packages=["inseminator", "inseminator.integrations"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)



from sqlalchemy import create_engine
from sqlalchemy.orm import Session, sessionmaker

from .configuration import Configuration


class Database:
    def __init__(self, configuration: Configuration) -> None:
        self.__engine = create_engine(configuration.connection_string)
        self.__session_maker = sessionmaker(self.__engine)

    def get_session(self) -> Session:
        return self.__session_maker()



from typing import Protocol

from inseminator import Container


class DomainModel(Protocol):
    def domain_logic(self, input_value: int) -> int:
        raise NotImplementedError


class Controller:
    def __init__(self, domain_model: DomainModel):
        self.__domain_model = domain_model

    def handler(self, input_value: int) -> int:
        return self.__domain_model.domain_logic(input_value)


# domain model implementation


class ConcreteDomainModel:
    def __init__(self):
        self.__logic_constant = 1

    def domain_logic(self, input_value: int) -> int:
        return input_value + self.__logic_constant


# entry point of you application

container = Container()

# view layer handling

controller = container.resolve(Controller, domain_model=ConcreteDomainModel())
result = controller.handler(1)
print(result)
